/* abc_test.c - test program for Adaptive Block Coder */
/* 01/19/2008, written by Yuriy A. Reznik <yreznik@ieee.org> */

/* This software and documentation is provided "as is", and the author makes 
 * no representations or warranties, expressed or implied, including, but not
 * limited to, warranties of merchantability or fitness for any particular
 * purpose, or that the use of this software or documentation will not infringe 
 * any third parties patents, copyrights, trademarks or other rights. The 
 * author will not be liable for any direct, indirect, special or consequential 
 * damages arising out of any use of the software or documentation. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "bitstream.h"
#include "abc.h"


#define M   100     /* max # of blocks in test sequence */
#define Q   10000   /* # of iterations */

/* test program: */
int main ()
{
  /* in/out buffers: */
  static unsigned char in_buffer [M*ABC_N/8];
  static unsigned char code_buffer [M*ABC_N/8 + 1024];
  static BITSTREAM in, code;
  
  /* vars: */
  unsigned int block, cx_blocks;
  unsigned int cx_counts[ABC_B+1], cx_count;
  int i, j, k, m; 
  double p, h, c;
  
  /* init ABC library: */
  abc_enc_init ();
  abc_dec_init ();
  
  /* scan sources: */
  for (p=0.5; p<=0.999; p+=0.01) {
    
    /* compute entropy: */
    h = - (p * log(p) + (1.-p) * log(1.-p)) / log(2.);
    printf ("\np=%g, h=%g\n", p, h);
    //printf("[");
    
    /* try different # of blocks: */
    for (m=1; m<M; m++) 
    {
      c = 0.;
      /* reset generator: */
      srand(1);
      /* make Q runs: */
      for (i=0; i<Q; i++) {                
        
        /* generate test sequence: */
        memset(in_buffer, 0, sizeof in_buffer);
        bitstream_open(&in, in_buffer, 0, 1);
        for (j=0; j<ABC_N*m; j++) {
          /* simulate Bernoulli source: */
          k = ((double) rand() / (double) RAND_MAX) > (1. - p);
          /* insert bit in the bitstream: */
          bitstream_write(&in, k, 1);
        }
        bitstream_close (&in);
        
        /*********** TEST ENCODING: ************************/
        
        /* open bitstreams: */
        bitstream_open(&in, in_buffer, 0, 0);
        memset(code_buffer, 0, sizeof code_buffer);   /* zero bits prior to writing */
        bitstream_open(&code, code_buffer, 0, 1);
        
        /* init context information: */
        for (k=0; k<=ABC_B; k++) cx_counts[k] = 0;
        cx_count = 0;

        /* run the encoder: */
        for (j=0; j<m; j++) {
          /* retrieve block to be encoded: */
          block = bitstream_read (&in, ABC_N);
          /* get # of blocks to be used as a context: */
          cx_blocks = min(j,ABC_B);
          /* encode block: */
          k = abc_enc(block, cx_blocks, cx_count, &code);
          /* update context: */
          cx_count += k - cx_counts[j % ABC_B];
          cx_counts[j % ABC_B] = k;                 /* save # of "1" in current block */
        }

        /* close bitstreams: */
        bitstream_close (&in);
        k = bitstream_close (&code);
        
        /* compute coding cost: */
        c += (double) k / (double)(m*ABC_N);
        
        /*********** TEST DECODING: ************************/
        
        /* open bistreams: */
        bitstream_open (&in, in_buffer, 0, 0);
        bitstream_open (&code, code_buffer, 0, 0);

        /* init context information: */
        for (k=0; k<=ABC_B; k++) cx_counts[k] = 0;
        cx_count = 0;

        /* run the decoder: */
        for (j=0; j<m; j++) {
          /* get # of blocks to be used as a context: */
          cx_blocks = min(j,ABC_B);
          /* decode block: */
          k = abc_dec(&block, cx_blocks, cx_count, &code);
          /* update context: */
          cx_count += k - cx_counts[j % ABC_B];
          cx_counts[j % ABC_B] = k;                 /* save # of "1" in current block */
          /* compare with the original block: */
          if (block != bitstream_read (&in, ABC_N)) {
            fprintf(stderr, "?%d,", j);
          }
        }

        /* close bitstreams: */
        bitstream_close (&in);
        bitstream_close (&code);
      }
      
      /* print results: */
      c /= (double)Q;
      printf("p=%g, H=%g, m=%d, C=%g, R=%g, ro=%g%%\n", p, h, m, c, c-h, 100.*(c-h)/h);
      //MAPLE plots:
      //printf("[%d,%g]%s", m*ABC_N, (c-h), (m<M-1)? ",": "],\n");
      fflush(stdout);
    }
  }
  
  return 1;
}
