/* bitstream.c - bitstream primitives */
/* 01/19/2008, written by Yuriy  A. Reznik <yreznik@ieee.org> */

/* This software and documentation is provided "as is", and the author makes 
 * no representations or warranties, expressed or implied, including, but not
 * limited to, warranties of merchantability or fitness for any particular
 * purpose, or that the use of this software or documentation will not infringe 
 * any third parties patents, copyrights, trademarks or other rights. The 
 * author will not be liable for any direct, indirect, special or consequential 
 * damages arising out of any use of the software or documentation. */

#include <stdio.h>
#include <limits.h>
#include "bitstream.h"

/* we'll need 32-bit or longer integers: */
#if UINT_MAX < 0xffffffff
#error "Integers are assumed to be (at least) 32-bit long!"
#endif

/* read/write 32-bits words from/to memory; 
 * pointer p is not word-aligned, so we do it byte by byte */
#define read_word(p,b)                \
  do {                                \
    b =  (unsigned) p[0] << 24;	      \
    b |= (unsigned) p[1] << 16;	      \
    b |= (unsigned) p[2] << 8;	      \
    b |= (unsigned) p[3];             \
  } while (0)

#define write_word(p,b)               \
  do {                                \
    p[0] = (unsigned char) (b >> 24); \
    p[1] = (unsigned char) (b >> 16); \
    p[2] = (unsigned char) (b >> 8);  \
    p[3] = (unsigned char) b;         \
  } while (0)


/* open bitstream for reading/writing: */
void bitstream_open (BITSTREAM *p, unsigned char *pp, int offset, int write)
{
  unsigned bb;
  /* preload bits in bit buffer: */
  if (write) bb = (unsigned) pp[0] << 24;	/* load last (partially filled) byte */
  else       read_word(pp,bb);
  /* save bitstream parameters: */
  p->start_p = p->p = pp;
  p->start_offset = p->offset = offset;
  p->bb = bb;
  p->mode = write;
}

/* close bitstream & return total number of bits read/written: */
unsigned int bitstream_close (BITSTREAM *p)
{
  unsigned char *pp = p->p;
  unsigned offset = p->offset;
  /* flush bit buffer: */
  if (p->mode) write_word(pp,p->bb);
  /* normalize & save bitstream position: */
  p->p = pp + (offset >> 3);
  p->offset = offset & 7;
  /* return number of bits read/written: */
  return (p->p - p->start_p) * 8 + p->offset - p->start_offset;
}

/* write "len" bits to the bitstream: */
void bitstream_write (BITSTREAM *p, unsigned bits, int len)
{
  int shift = 32 - len;
  unsigned offset = p->offset;
  /* check if bit buffer is full: */
  if ((shift -= offset) <= 0) {
    unsigned char *pp = p->p;
    unsigned bb = p->bb, written;
    /* flush bit buffer: */
    write_word(pp,bb);
    /* normalize position: */
    p->p = pp + (offset >> 3);
    written = offset & ~7; offset &= 7;
    p->bb = bb << written;
    shift += written;
  }
  /* append new bits: */
  p->bb |= bits << shift;
  p->offset = offset + len;
}

/* shifts the current position in the bitstream: */
void bitstream_scroll (BITSTREAM *p, int len)
{
  unsigned offset = p->offset + len;
  unsigned bb = p->bb << len;
  /* shall we reload bit buffer? */
  if (offset > 32 - BITSTREAM_MAX_LEN) {
    unsigned char *pp = p->p;
    /* normalize position/offset: */
    pp += offset >> 3; offset &= 7; p->p = pp;
    /* reload bit buffer: */
    read_word(pp,bb);
    bb <<= offset;
  }
  /* save updated bitstream vars: */
  p->offset = offset;
  p->bb = bb;
}

/* read "len" bits from the bitstream: */
unsigned bitstream_read (BITSTREAM *p, int len)
{
  unsigned bits = bitstream_buffer(p) >> (32-len);
  bitstream_scroll (p, len);
  return bits;
}

/* bitstream.c -- end of file */